// PDEsim.cpp: implementation of the PDEsim class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "FourierPDE.h"
#include "PDEsim.h"
extern "C"
{
#include "fft/fft.h"
}

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

PDEsim::PDEsim()
{

	// Default values
	dislocationSeparation = 0;
	dt = 0.5;
	kx = 0.;
	ky = 0.;
	noise = 0.1;
	SetNu(7);
	seed = 1;
	min = max = 0;
	psi = 0;	// Allocate in Restart();
	expLdt = 0;	// Allocate in Restart();

	// Swift-Hohenberg
	epsilon = .4;
	k = 1.;
	Nk = NkOld = 0;
	expNonLinear = 0;

	// Generalized Swift-Hohenberg
	g_m = 50.;
	c2 = 2.0;
	zeta = 0;
	meanLin = 0;
	// Internal to GSH computations
	dDel2PsiDx = 0;
	dDel2PsiDy = 0;
	dPsiDx = 0;
	dPsiDy = 0;
	dZetaDx = 0;
	dZetaDy = 0;

	Restart();

}

PDEsim::~PDEsim()
{
		delete [] psi;
		delete [] expLdt;

		// Swift-Hohenberg
		delete [] Nk;
		delete [] NkOld;
		delete [] expNonLinear;

		// Generalized Swift-Hohenberg
		delete [] zeta;
		delete [] dDel2PsiDx;
		delete [] dDel2PsiDy;
		delete [] dPsiDx;
		delete [] dPsiDy;
		delete [] dZetaDx;
		delete [] dZetaDy;
		delete [] meanLin;
}


double PDEsim::Sweep()
{
	// Swift-Hohenberg, GSH
	// Calculate nonlinear part for Psi
	ComputeNk(); 

	fftn(psi,2,dim);

	for (int i=0; i<dim[0]*dim[1]; i++)
	{
		psi[i][0] = expLdt[i]*psi[i][0]
			+ expNonLinear[i]*(1.5*Nk[i][0] - 0.5*NkOld[i][0]);	// SH, GSH
		psi[i][1] = expLdt[i]*psi[i][1]
			+ expNonLinear[i]*(1.5*Nk[i][1] - 0.5*NkOld[i][1]);	// SH, GSH
	}

// SH, GSH
// Swap Nk and NkOld
	myComplex* NkHoldPointer = NkOld;
	NkOld = Nk;
	Nk = NkHoldPointer;

// Generalized Swift-Hohenberg
	ComputeZetas();

	invfftn(psi,2,dim);
	normalize_fftn(psi,2,dim);

	// Bruce Roberts found this instability: small complex pieces of psi blow up
	for (i=0; i<dim[0]*dim[1]; i++)
	{
		psi[i][1] = 0.0;
	}

	return 1;

}

void PDEsim::SetRandom()
{
	// Dislocation
	double x0 = 0.5*(dim[0]-dislocationSeparation);
	double x1 = 0.5*(dim[0]+dislocationSeparation);
	double y = 0.5*dim[1];
	// Set psi to real values
	for (int i=0; i<dim[0]; i++)
	{
		for (int j=0; j<dim[1]; j++)
		{
			psi[i*dim[0]+j][0] = (2.0*rand.uniform()-1.0)*noise 
				+ sin(kx*i + ky*j + atan2((j-y),(i-x0)) - atan2((j-y),(i-x1)));
			psi[i*dim[0]+j][1] = 0;
		}
	}
}


// SH, GSH

void PDEsim::ComputeNk()
{


	for(int dub=0; dub<dim[0]*dim[1]; dub++)
	{
		Nk[dub][0] = -psi[dub][0]*psi[dub][0]*psi[dub][0]
			- g_m * (dZetaDy[dub][0]*dPsiDx[dub][0]-dZetaDx[dub][0]*dPsiDy[dub][0]); // Generalized Swift-Hohenberg
		Nk[dub][1] = 0.0;
	}

	fftn(Nk, 2, dim);

}


void PDEsim::Restart()
{
	// psi is 2^nu x 2^nu array of complex numbers
	
	if (psi) delete [] psi;
	if (expLdt) delete [] expLdt;

	// Swift-Hohenberg
	if (Nk) delete [] Nk;
	if (NkOld) delete [] NkOld;
	if (expNonLinear) delete [] expNonLinear;


	// Generalized Swift-Hohenberg
	if (zeta) delete [] zeta;
	if (meanLin) delete [] meanLin;
	if (dDel2PsiDx) delete [] dDel2PsiDx;
	if (dDel2PsiDy) delete [] dDel2PsiDy;
	if (dPsiDx) delete [] dPsiDx;
	if (dPsiDy) delete [] dPsiDy;
	if (dZetaDx) delete [] dZetaDx;
	if (dZetaDy) delete [] dZetaDy;


	psi = new myComplex[dim[0]*dim[1]];
	expLdt = new double[dim[0]*dim[1]];

	// Swift-Hohenberg
	Nk = new myComplex[dim[0]*dim[1]];
	NkOld = new myComplex[dim[0]*dim[1]];
	expNonLinear = new double[dim[0]*dim[1]];


	// Generalized Swift-Hohenberg
	zeta = new myComplex[dim[0]*dim[1]];
	meanLin = new double[dim[0]*dim[1]];
	dDel2PsiDx = new myComplex[dim[0]*dim[1]];
	dDel2PsiDy = new myComplex[dim[0]*dim[1]];
	dPsiDx = new myComplex[dim[0]*dim[1]];
	dPsiDy = new myComplex[dim[0]*dim[1]];
	dZetaDx = new myComplex[dim[0]*dim[1]];
	dZetaDy = new myComplex[dim[0]*dim[1]];

	rand.seed(seed);

	SetRandom();

// SH, GSH
	ComputeNk();
// Swap Nk and NkOld
	myComplex* NkHoldPointer = NkOld;
	NkOld = Nk;
	Nk = NkHoldPointer;


	double qx, qy;
	double dq = (2 * 3.1415928)/dim[0];	// Separate dqx and dqy if not square!
	for (int i=0; i<dim[0]; i++)
	{
		if (i < dim[0]/2)	qx = i*dq;
		else				qx = (i-dim[0])*dq;
		for (int j=0; j<dim[1]; j++)
		{
			if (j < dim[1]/2)	qy = j*dq;
			else				qy = (j - dim[1])*dq;

			double q2 = qx*qx+qy*qy;

// diffusion equation	
//			expLdt[i*dim[0]+j] = exp(- q2 * dt);

// Swift-Hohenberg
			
			double Lk = epsilon - (k*k-q2)*(k*k-q2);
			expLdt[i*dim[0]+j] = exp(Lk * dt);
			expNonLinear[i*dim[0]+j] = (expLdt[i*dim[0]+j] - 1.0)/Lk;
			if (q2!=0) meanLin[i*dim[0]+j] = -1.0/((q2+c2)*q2);	// Generalized Swift-Hohenberg
			else meanLin[i*dim[0]+j]=0;	// GSH; Mean flow has mean zero

		}
	}

// Generalized Swift-Hohenberg: ComputeZetas needs Psi(k)
	fftn(psi,2,dim);
	ComputeZetas();
	invfftn(psi,2,dim);
	normalize_fftn(psi,2,dim);

}

// Generalized Swift-Hohenberg
void PDEsim::ComputeZetas()
{
	// Given current Psi in k-space calculate zeta, grad Psi, and grad Zeta, in real space

	double qx, qy;
	double dq = (2 * 3.1415928)/dim[0];	// Separate dqx and dqy if not square!

	for (int i=0; i<dim[0]; i++)
	{
		if (i < dim[0]/2)	qx = i*dq;
		else				qx = (i-dim[0])*dq;
		for (int j=0; j<dim[0]; j++)
		{
			int dub = i*dim[0]+j;
			if (j < dim[1]/2)	qy = j*dq;
			else				qy = (j - dim[1])*dq;

			double q2 = qx*qx+qy*qy;

			dPsiDx[dub][1] =      qx * psi[dub][0];
			dPsiDx[dub][0] =     -qx * psi[dub][1];
			dPsiDy[dub][1] =      qy * psi[dub][0];
			dPsiDy[dub][0] =     -qy * psi[dub][1];
			dDel2PsiDx[dub][1] =  qx * (-q2) * psi[dub][0];
			dDel2PsiDx[dub][0] = -qx * (-q2) * psi[dub][1];
			dDel2PsiDy[dub][1] =  qy * (-q2) * psi[dub][0];
			dDel2PsiDy[dub][0] = -qy * (-q2) * psi[dub][1];

		}
	}

	invfftn(dPsiDx,2,dim);
	invfftn(dPsiDy,2,dim);
	invfftn(dDel2PsiDx,2,dim);
	invfftn(dDel2PsiDy,2,dim);
	normalize_fftn(dPsiDx,2,dim);
	normalize_fftn(dPsiDy,2,dim);
	normalize_fftn(dDel2PsiDx,2,dim);
	normalize_fftn(dDel2PsiDy,2,dim);

	// Generalized Swift-Hohenberg
	// Calculating zeta: use zeta array as temporary storage for Mk

	for (i=0; i<dim[0]; i++)
	{
		for (int j=0; j<dim[0]; j++)
		{
			int dub = i*dim[0]+j;
			zeta[dub][0] = (dDel2PsiDx[dub][0]*dPsiDy[dub][0] - dDel2PsiDy[dub][0]*dPsiDx[dub][0]);
			zeta[dub][1] = 0.0;
		}
	}
	fftn(zeta,2,dim);	// Mk

	for (i=0; i<dim[0]; i++)
	{
		if (i < dim[0]/2)	qx = i*dq;
		else				qx = (i-dim[0])*dq;
		for (int j=0; j<dim[0]; j++)
		{
			int dub = i*dim[0]+j;
			if (j < dim[1]/2)	qy = j*dq;
			else				qy = (j - dim[1])*dq;

			zeta[dub][0] *= meanLin[dub];	// Convert from Mk to zeta
			zeta[dub][1] *= meanLin[dub];

			dZetaDx[dub][1] =      qx * zeta[dub][0];
			dZetaDx[dub][0] =     -qx * zeta[dub][1];
			dZetaDy[dub][1] =      qy * zeta[dub][0];
			dZetaDy[dub][0] =     -qy * zeta[dub][1];
		}
	}

	invfftn(zeta,2,dim);	// Zeta is computed here only for plotting

	invfftn(dZetaDx,2,dim);	// Used to calculate Nk
	invfftn(dZetaDy,2,dim);

	normalize_fftn(zeta,2,dim);
	normalize_fftn(dZetaDx,2,dim);
	normalize_fftn(dZetaDy,2,dim);

}
